#!/usr/bin/ruby
require 'hdatastructures/hlist'
require 'hmisc/hcolorize'

class HFieldTable < Array

  attr_accessor :tableName, :allRows, :allCols

  def initialize(allRows = false, allCols = false)
    @tableName = nil
    @fields = {}
    @allRows = allRows
    @allCols = allCols
  end

  def copyConstructor
    Marshal.load(Marshal.dump(self))
  end

  def fields()
    result = {}
    @fields.each do |fieldName, fieldValue|
      result[fieldName] = fieldValue if fieldValue.visible or @allCols
    end
    return result
  end
  
  def addFieldName(fieldName, visible = true)
    record = HRecord.new(fieldName) unless @fields[fieldName]
    record.set(:visible, visible)
    @fields[fieldName] = record
  end
 
  def method_missing(methodName, *args)

    if methodName.to_s =~ /^setField(.+)$/ # example: setFieldCaption(fieldName, fieldCaption)
      @fields[args[0]].set($1.downcase().to_sym, args[1])
    else
      caller_infos = caller.first.split(":")
      puts "Error: #{caller_infos[0]}:#{caller_infos[1]} - #{methodName} not found".hight_red()  
    end

  end

  def rowCount()
    return self.length()
  end

  def columnCount()
    return @fields.length()
  end
  
  def columnCountByKey(key = :visible)
    return self.columnCount if @allCols
    i = 0
    @fields.each { |fieldName, fieldValue| i += 1 if fieldValue.value(key) }
    return i
  end

  def setData(row, fieldName, data)
    self[row][fieldName] = data
  end

  # Inserts a new row filled with data
  def insertRow(row, data = nil)
    tmp = @fields.dup
    self.insert(row, tmp.each { |k, v| tmp[k] = data } )
  end

  def deleteRow(row)
    self.delete_at(row)
  end

  def setDataByFieldName(row, fieldName, data)
    self.setData(row, fieldName, data) 
  end
  
  def setIntoRecordByFieldName(row, fieldName, data)
    self.setDataByFieldName(row, fieldName, HRecord.new(data)) 
  end

  def setCaption
    self.fields.each do |fieldName, fieldValue|
      self.setFieldCaption(fieldName, fieldValue.value.hcapitalize())
    end
  end

  def [](fieldName)
    self[fieldName] = {} unless super(fieldName)
    return super(fieldName)
  end

  def data(row, fieldName)
    return self[row][fieldName]
  end
  
  def dataByFieldName(row, fieldName)
    return self.data(row, fieldName)
  end
  
  def valueByFieldName(row, fieldName)
    value = self.data(row, fieldName)
    value = value.value() if (value.class == HRecord)
    return value
  end
  
  def row(row)
    #return self[row] if @allCols
    result = {}
    @fields.each do |fieldName, fieldValue|
      result[fieldName] = self.dataByFieldName(row, fieldName) if fieldValue.visible or @allCols
    end
    return result
  end

  def match(row)
    @fields.each do |fieldName, fieldValue|
      filter = fieldValue.filter
      next unless filter
      data = self.dataByFieldName(row, fieldName)
      return false unless data
      data = HRecord.new(data) if data.class != HRecord
      filter = HRecord.new(filter) if filter.class != HRecord
      return false if data != filter
    end
    return true

  end

  def each()
    for i in 0...self.rowCount() do
      row = self.row(i)
      yield(row, i) if @allRows or self.match(i) 
    end 
  end

  def rowWithFieldName(row)
    result = {}
    self[row].each do |fieldName, fieldValue|
      field = @fields[fieldName]
      result[fieldName] = fieldValue if field.visible or @allCols
    end
    return result
  end
  
  def eachWithFieldName()
    for i in 0...self.rowCount() do
      row = self.rowWithFieldName(i)
      yield(row, i) if @allRows or self.match(i) 
    end 
  end
  
  
  def sumByFieldName(fieldName)

    total = 0
    for i in 0...self.rowCount() do
      total += (self.dataByFieldName(i, fieldName)).to_f
    end 
    return total
  
  end

  def valueByType(fieldName, fieldValue)

    return nil unless fieldValue  
    type = @fields[fieldName].value(:type) if @fields[fieldName]
    value = fieldValue.value 
    return value unless type
    return value.to_s if type == "text"
    return value.to_i if type == "integer"
    return value.to_f if type == "float"
    return value.to_date() if type == "date"
    return value.to_datetime() if type == "datetime"
    return value.to_s

  end

  # fieldTable.sortBy("price")
  # fieldTable.sortBy("price", "amount")
  # fieldTable.sortBy("price", "amount", asc: false)
  def sortBy(*fieldNames, asc: true)
    #self.herbertsort!(asc) { |row| fieldNames.map {|fieldName| row[fieldName]} } 
    self.herbertsort!(asc) { |row| fieldNames.map {|fieldName| self.valueByType(fieldName, row[fieldName])} } 
  end
  
  def showHeader(key = :key)
    self.fields.each do |fieldName, fieldValue|
      print $hpformat % fieldValue.value(key) 
    end
    puts
  end

  def showData(data, key = :key)

    if !data
      print $hpformat % "." unless data
    elsif data.class == Fixnum or data.class == String
      print $hpformat % data
    else
      data.show(key)
    end
  
  end

  def showTotalizers()

    @fields.each do |fieldName, fieldValue|
      totalizerName = fieldValue.totalizer
      if totalizerName 
        print $hpformat % totalizerName
        print $hpformat % self.sumByFieldName(fieldName)
        puts
      end
    end
    puts

  end


  def show(key = :key)
    print "=============>  HFieldTable  <==============\n"
    self.showHeader(:caption)
    self.each do |row, i|
      row.each do |fieldName, fieldValue|    
        self.showData(fieldValue)
      end
      puts
    end
    puts
  end
  

  def self.test()

    table = HFieldTable.new()
    table.setData(0, 0, HRecord.new("0x0"))
    table.setData(10, 1, HRecord.new("10x1"))
    table.setData(9, 9, HRecord.new("9x9"))
    table.show()
  end

  def self.test2()

    fieldTable = HFieldTable.new()
    fieldTable.addFieldName("quantity")
    fieldTable.addFieldName("description")
    fieldTable.addFieldName("price", false)
    fieldTable.addFieldName("amount")

    fieldTable.setFieldCaption("quantity", "Quantity")
    fieldTable.setFieldCaption("description", "Description")
    fieldTable.setFieldCaption("price", "Price")
    fieldTable.setFieldCaption("amount", "Amount")

    fieldTable.setFieldFilter("price", HRecord.new(10))
    fieldTable.setFieldFilter("amount", HRecord.new(10))

    for i in 0..30
      fieldTable.setDataByFieldName(i, "quantity", HRecord.new(i))
      fieldTable.setDataByFieldName(i, "description", HRecord.new(i))
      fieldTable.setDataByFieldName(i, "price", HRecord.new(rand(10)))
      fieldTable.setDataByFieldName(i, "amount", HRecord.new(i))
    end

    #fieldTable.insertRow(1, HRecord.new("ciao"))
    #fieldTable.deleteRow(1)

    fieldTable.show()
    fieldTable.allRows,fieldTable.allCols = true, true
    fieldTable.show()

    fieldTable.sortBy("price", "amount", asc: false)
    fieldTable.show()

    return fieldTable

  end


end

t1 = Time.now
HFieldTable.test2()
p "... msecs. #{((Time.now - t1) * 1000).round(2)}"
